//matrix.cpp
#include "matrix.h"
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <cmath>

using std::cout; using std::endl;

//Constructors/destructors:

//constructor from pointer-to-pointer 
Matrix::Matrix(int r, int c, double** pa){
  nRows = r;
  nCols = c;
  pA = new double*[nRows];
  memoryCheck();
  
  for(int i = 0; i < nRows; i++){
    pA[i] = new double[nCols];
    memoryCheck(i);
  }

  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < nCols; j++)
      pA[i][j] = pa[i][j];
}


//diagonal matrix constructor
Matrix::Matrix(int r, int c, double a){
  nRows = r;
  nCols = c;
  
  //set pA to null pointer in default case
  if(r == 0 || c == 0)
    pA = 0;

  pA = new double*[r];
  memoryCheck();
      
  for(int i = 0; i < nRows; i++){
    pA[i] = new double[nCols];
    memoryCheck(i);
  }

  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < nCols; j++){
      if(i == j)
        pA[i][j] = a;
      else
        pA[i][j] = 0.;
    }
}

//copy constructor
Matrix::Matrix(const Matrix& A){
  nRows = A.nRows;
  nCols = A.nCols;
  pA = new double*[nRows];
  memoryCheck();
      
  for(int i = 0; i < nRows; i++){
    pA[i] = new double[nCols];
    memoryCheck(i);
  }
  
  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < nCols; j++)
      pA[i][j] = A.pA[i][j];
}    
      
//destructor
Matrix::~Matrix(){   
  if(pA != 0){
    for(int i = 0; i < nRows; i++)
      if(pA[i] != 0)
        delete[] pA[i];
    delete[] pA;
  }
}
    
//Overloaded operators:

Matrix& Matrix::operator=(const Matrix& A){

  if(this == &A) //avoid self assignment
    return *this;

  if(pA != 0){
    for(int i = 0; i < nRows; i++) //release current memory
      delete[] pA[i];
    delete[] pA;
  }
    
  //define/redefine object's members
  nRows = A.nRows; 
  nCols = A.nCols;
  pA = new double*[nRows];
  memoryCheck();
      
  for(int i = 0; i < nRows; i++){
    pA[i] = new double[nCols];
    memoryCheck(i);
  }

  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < nCols; j++)
      pA[i][j] = A.pA[i][j];

  return *this;
}    

Matrix Matrix::operator+(const Matrix& B){
  if(nRows != B.nRows || nCols != B.nCols){
    cout << "Bad row and/or column dimensions in +!" << endl;
    exit(1);
  }
  Matrix temp(nRows, nCols);
  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < nCols; j++)
      temp.pA[i][j]=this->pA[i][j] + B.pA[i][j];
  
  return temp;
}

Matrix& Matrix::operator+=(const Matrix& B){
  if(nRows != B.nRows || nCols != B.nCols){
    cout << "Bad row and/or column dimensions in +=!" << endl;
    exit(1);
  }
  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < nCols; j++)
      pA[i][j] += B.pA[i][j];
  
  return *this;
}

Matrix Matrix::operator-(const Matrix& B){
  if(nRows != B.nRows || nCols != B.nCols){
    cout << "Bad row and/or column dimensions in -!" << endl;
    exit(1);
  }
  Matrix temp(nRows, nCols);
  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < nCols; j++)
      temp.pA[i][j]=this->pA[i][j] - B.pA[i][j];

  return temp;
}

Matrix& Matrix::operator-=(const Matrix& B){
  if(nRows != B.nRows || nCols != B.nCols){
    cout << "Bad row and/or column dimensions in -=!" << endl;
    exit(1);
  }
  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < nCols; j++)
      pA[i][j] -= B.pA[i][j];

  return *this;
}

Matrix Matrix::operator*(const Matrix& B){
  if(nCols != B.nRows){
    cout << "Bad row and column dimensions in *!" << endl;
    exit(1);
  }
  Matrix C(nRows, B.nCols); // matrix of all 0s
  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < B.nCols; j++)
      for(int k = 0; k < B.nRows; k++)
        C.pA[i][j] += pA[i][k]*B.pA[k][j];
  
  return C;
}

Matrix Matrix::operator*(double b){
  Matrix C(nRows, nCols); // matrix of all 0s
  for(int i = 0; i < nRows; i++)
    for(int j = 0; j < nCols; j++)
      C.pA[i][j] += b*pA[i][j];
  
  return C;
}

Vector Matrix::operator*(Vector v){
  double temp;
  Vector c(nRows); // matrix of all 0s
  for(int i = 0; i < nRows; i++){
    temp = 0;
    for(int j = 0; j < nCols; j++)
      temp += pA[i][j]*v[j];
    
    c.pA[i]=temp;
  }
  
  return c;
}

//matrix operations

Matrix Matrix::trans(){
  Matrix B(nCols, nRows, 0.);
  for(int i = 0; i < nCols; i++)
    for(int j = 0; j < nRows; j++)
      B.pA[i][j] = pA[j][i];
  return B;
}


//solution of a linear system

//back-solve an upper triangular system 
Vector Matrix::backward(Vector& RHS){
  double temp; //temporary storage
  
  Vector b(nRows, 0.);//solution vector
  
  //initialize recursion on last row of all right-hand-sides
  if(pA[nRows-1][nRows-1] != 0)
    b.pA[nRows - 1] = RHS.pA[nRows - 1]/pA[nRows - 1][nRows - 1];
  else{
    cout << "Singular system!" << endl;
    exit(1);
  }
  
  //now work through the remaining rows
  for(int i = (nRows - 2); i >= 0; i--){
    if(pA[i][i] != 0){
      temp = RHS.pA[i];
      for(int k = (i+1); k < nRows; k++)
        temp -= b.pA[k]*pA[i][k]; 
      b.pA[i] = temp/pA[i][i];
    }
    else{
      cout << "Singular system!" << endl;
      exit(1);
    }
  }
  return b;
}

//forward-solve a lower triangular system 
Vector Matrix::forward(Vector& RHS){
  double temp; //temporary storage
  Vector b(nRows, 0.);//solution matrix

  //initialize recursion on first row of the right-hand-side
  if(pA[0][0] != 0)
    b.pA[0] = RHS.pA[0]/pA[0][0];
  else{
    cout << "Singular system!" << endl;
    exit(1);
  }
  
  //now work through the remaining rows
  for(int i = 1; i < nRows; i++){
    if(pA[i][i] != 0){
      temp = RHS.pA[i];
      for(int k = 0; k < i; k++)
        temp -= b.pA[k]*pA[i][k]; 
      b.pA[i] = temp/pA[i][i];
    }
    else{
      cout << "Singular system!" << endl;
      exit(1);
    }
  }
  return b;
}

Matrix Matrix::cholInv(){
  double temp = 0;
  Matrix G(nRows, nRows, 0.);

  for(int j = 0; j < nRows; j++){//proceed by columns
    if(pA[j][j] == 0){
      cout << "Singular system!" << endl;
      exit(1);
    }
    temp = pA[j][j]; //starting value for diagonal element recursion
    
    if(j > 0)
      for(int k = 0; k < j; k++)
        temp -= G[j][k]*G[j][k];
    
    G.pA[j][j] = sqrt(temp);
    
    for(int i = (j + 1); i < nRows; i++){//now do the rest of column j
      temp = pA[j][i];
      for(int k = 0; k < j; k++)
        temp -= G[j][k]*G[i][k];
      G.pA[i][j] = temp/G[j][j];
    }
  }

  Vector temp1, temp2;
  Matrix Ainv(nRows, nRows, 0.);
  for(int i = 0; i < nRows; i++){
    temp1 = Vector(nRows, i);
    temp2 = G.forward(temp1);
    temp1 = G.trans().backward(temp2);
    Ainv.vecToMat(i, temp1);
  }

  return Ainv;
}    

// Utilities:

void Matrix::printMatrix(){
  for(int i = 0; i < nRows; i++){
    for(int j = 0; j < nCols; j++)
      cout << " " << std::setw(10) << std::setprecision(7)  
           << pA[i][j] << "  ";
    cout << endl;
  }
}

Vector Matrix::matToVec(int j){
  Vector v(nRows, 0.);
  for(int i = 0; i < nRows; i++)
    v.pA[i] = pA[i][j];
  return v;
}

void Matrix::vecToMat(int j, Vector& v){
  for(int i = 0; i < nRows; i++)
    pA[i][j] = v[i];
}

void Matrix::memoryCheck(){
  if(pA == 0){
    cout << "Memory allocation for pA failed" << endl;
    exit(1);
  }
}

void Matrix::memoryCheck(int i){
  if(pA[i] == 0){
    cout << "Memory allocation for pA[" << i << "] failed" << endl;
    exit(1);
  }
}

//non-member operators

Matrix operator*(Vector& v1, Vector& v2){
  Matrix B(v1.getNRows(), v2.getNRows());
  for(int i = 0; i < v1.getNRows(); i++)
    for(int j = 0; j < v2.getNRows(); j++)
      B.pA[i][j] = v1[i]*v2[j];
  return B;
}

Matrix operator*(double b, const Matrix& A){
  Matrix B = Matrix(A.nRows, A.nCols);
  for(int i = 0; i < A.nRows; i++)
    for(int j = 0; j < A.nCols; j++)
      B.pA[i][j] = b*A.pA[i][j];
  return B;
}
